/**************************************************************************************************
Copyright (C) 2016 Andrey Tsyganov, UlSPU. All rights reserved.

Contact information
-------------------
Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
Web    :  http://lmm.ulspu.ru
e-mail :  andrew.tsyganov@gmail.com
**************************************************************************************************/


#include <iostream>
#include <ostream>
#include <conio.h>
#include <math.h>
#include "../../contrib/randomc/randomc.h"
#include "../../src/ga/ga_seq.h"


const int    dim = 20; // Number of dimensions
const double lbound = -5.12; // Lower bound for each coordinate
const double ubound = 5.12; // Upper bound for each coordinate
const double max_bias = ubound; // Maximum bias of the coordinate


#ifndef M_PI
// Source: http://www.geom.uiuc.edu/~huberty/math5337/groupe/digits.html
#define M_PI 3.141592653589793238462643383279502884197169399375105820974944592307816406 
#endif


// some constants to speedup calculations
const double k = 2 * M_PI; // constant 2 * Pi coefficient in cost function
const double s0 = 10 * dim; // constant 10 * dim term in cost fuction


class Indiv
{
	friend class Rastrigin_GA_SEQ;
	friend std::ostream& operator<<(std::ostream& os, const Indiv& indiv)
	{
		os << "(";
		for (int i = 0; i < dim; i++)
		{
			os << indiv.x_[i];
			if (i < dim - 1)
				os << "; ";
		}
		os << ")";

		return os;
	}

private:
	double x_[dim]; // coordinates
};


class Rastrigin_GA_SEQ: public ga::GA_SEQ< double, Indiv >
{
public:
	virtual void output(const ga::GA_Param<double>& param, const ga::GA_State<double, Indiv>& state, std::ostream& os);
	virtual double fitness(const Indiv& indiv, const ga::GA_State<double, Indiv>& state); // Evaluate fitness
	virtual Indiv generate(const ga::GA_State<double, Indiv>& state, ga::RndGen& rnd); // Generate individual
	virtual Indiv crossover(const Indiv& parent1, const Indiv& parent2, const ga::GA_State<double, Indiv>& state, ga::RndGen& rnd); // Perform crossover
	virtual Indiv mutation(const Indiv& indiv, const ga::GA_State<double, Indiv>& state, ga::RndGen& rnd); // Perform mutation 
};


void Rastrigin_GA_SEQ::output(const ga::GA_Param<double>& param, const ga::GA_State<double, Indiv>& state, std::ostream& os)
{
	if (state.gen_ == 0)
		os << state << "\n";
	else
	{
		/*
		os << "gen_: " << state.gen_ << "\n";
		os << "population_:\n";
		for (size_t i = 0; i < state.population_.size(); i++)
		os << std::setw(2) << i << ") " << state.population_[i] << " fitness_: " << state.fitnesses_[i] << "\n";
		*/
		os << "gen_: " << state.gen_ << " population_[0]: " << state.population_[0] << " fitnesses_[0]: " << state.fitnesses_[0] << "\n";
	}
}


double Rastrigin_GA_SEQ::fitness(const Indiv& indiv, const ga::GA_State<double, Indiv>& state)
{
	double sum = s0;

	for (int i = 0; i < dim; i++)
		sum += indiv.x_[i] * indiv.x_[i] - 10 * cos(k * indiv.x_[i]);

	return sum;
}


Indiv Rastrigin_GA_SEQ::generate(const ga::GA_State<double, Indiv>& state, ga::RndGen& rnd)
{
	Indiv indiv;

	for (int i = 0; i < dim; i++)
		indiv.x_[i] = lbound + (ubound - lbound) * rnd.Random();

	return indiv;
}


Indiv Rastrigin_GA_SEQ::crossover(const Indiv& parent1, const Indiv& parent2, const ga::GA_State<double, Indiv>& state, ga::RndGen& rnd)
{
	// Scattered crossover
	Indiv child;

	for (int i = 0; i < dim; i++)
		// Better (first) parent contributes two times more genes
		child.x_[i] = !(rnd.IRandom(0, 2)) ? parent2.x_[i] : parent1.x_[i];

	return child;
}


Indiv Rastrigin_GA_SEQ::mutation(const Indiv& indiv, const ga::GA_State<double, Indiv>& state, ga::RndGen& rnd)
{
	Indiv child = indiv;

	int coord = rnd.IRandom(0, dim - 1);

	for (int i = 0; i < dim; i++)
	{
		if (i == coord)
		{
			child.x_[i] += -max_bias + 2 * max_bias * rnd.Random();
		}
		else
		{
			double alpha = rnd.Random();
			child.x_[i] = alpha * child.x_[i] + (1 - alpha) * state.population_[0].x_[i];
		}

		if (child.x_[i] < lbound)
			child.x_[i] = lbound;

		if (child.x_[i] > ubound)
			child.x_[i] = ubound;
	}

	return child;
}


int main()
{
	ga::GA_Param<double> param;
	ga::GA_Result <double, Indiv> res;
	std::vector<Indiv> v_init_sol;
	Rastrigin_GA_SEQ ga;

	double start_time, end_time;

	param.setGoal(ga::MIN);
	param.setPopSize(15);
	param.setEliteCount(2);
	param.setDeadCount(2);
	param.setMigrationCount(3);
	param.setCrossoverFrac(0.6);

	//param.setStopCond(ga::GEN_LIM);
	//param.setGenLim(500);

	//param.setStopCond(ga::FITNESS_LIM);
	//param.setFitnessLim(0.1);

	param.setStopCond(ga::TIME_LIM);
	param.setTimeLim(1.5);

	param.setOfname("log.txt");

	std::cout << param << "\n";
	std::cout << "Start GA...\n";

	start_time = get_wall_time();
	ga.run(param, v_init_sol, res);
	end_time = get_wall_time();

	std::cout << "End GA\n";
	std::cout << "Elapsed time: " << end_time - start_time << " second(s)\n\n";
	std::cout << res << "\n";

	_getch();

	return 0;
}