/**************************************************************************************************
Copyright (C) 2016 Andrey Tsyganov, UlSPU. All rights reserved.

Contact information
-------------------
Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
Web    :  http://lmm.ulspu.ru
e-mail :  andrew.tsyganov@gmail.com
**************************************************************************************************/



#include <iostream>
#include <ostream>
#include <conio.h>
#include <math.h>
#include "../../contrib/randomc/randomc.h"
#include "../../src/sa/sa_omp.h"


const int    dim = 500; // Number of dimensions
const double lbound = -5.12; // Lower bound for each coordinate
const double ubound = 5.12; // Upper bound for each coordinate
const double max_bias = ubound; // Maximum bias of the coordinate


#ifndef M_PI
								// Source: http://www.geom.uiuc.edu/~huberty/math5337/groupe/digits.html
#define M_PI 3.141592653589793238462643383279502884197169399375105820974944592307816406 
#endif

								// some constants to speedup calculations
const double k = 2 * M_PI; // constant 2 * Pi coefficient in cost function
const double s0 = 10 * dim; // constant 10 * dim term in cost fuction


class Solution
{
	friend class Rastrigin_SA_OMP;
	friend std::ostream& operator<<(std::ostream& os, const Solution& solution)
	{
		os << "(";
		for (int i = 0; i < dim; i++)
		{
			os << solution.x_[i];
			if (i < dim - 1)
				os << "; ";
		}
		os << ")";

		return os;
	}

private:
	double x_[dim]; // coordinates
};


class Rastrigin_SA_OMP : public sa::SA_OMP< double, Solution >
{
public:
	virtual void output(const sa::SA_Param<double>& param, const sa::SA_State<double, Solution>& state, std::ostream& os);
	virtual double cost(const Solution& solution, const sa::SA_State<double, Solution>& state);
	virtual Solution generate(const sa::SA_State<double, Solution>& state, sa::RndGen& rnd);
	virtual Solution neighbour(const Solution& solution, const sa::SA_State<double, Solution>& state, sa::RndGen& rnd);
};


void Rastrigin_SA_OMP::output(const sa::SA_Param<double>& param, const sa::SA_State<double, Solution>& state, std::ostream& os)
{
	/*
	if (state.step_ == 0)
		os << state << "\n";
	else
		os << "step_: " << state.step_ << " stall_step_: " << state.stall_step_ << " best_:  " << state.best_ << " best_cost_: " << state.best_cost_ << "\n";
	*/
}


double Rastrigin_SA_OMP::cost(const Solution& solution, const sa::SA_State<double, Solution>& state)
{
	double sum = s0;

	for (int i = 0; i < dim; i++)
		sum += solution.x_[i] * solution.x_[i] - 10 * cos(k * solution.x_[i]);

	return sum;
}


Solution Rastrigin_SA_OMP::generate(const sa::SA_State<double, Solution>& state, sa::RndGen& rnd)
{
	Solution solution;

	for (int i = 0; i < dim; i++)
		solution.x_[i] = lbound + (ubound - lbound) * rnd.Random();

	return solution;
}


Solution Rastrigin_SA_OMP::neighbour(const Solution& solution, const sa::SA_State<double, Solution>& state, sa::RndGen& rnd)
{
	Solution neighb = solution;

	int coord = rnd.IRandom(0, dim - 1);

	for (int i = 0; i < dim; i++)
	{
		if (i == coord)
		{
			neighb.x_[i] += -max_bias + 2 * max_bias * rnd.Random();
		}
		else
		{
			double alpha = rnd.Random();
			neighb.x_[i] = alpha * neighb.x_[i] + (1 - alpha) * state.best_.x_[i];
		}

		if (neighb.x_[i] < lbound)
			neighb.x_[i] = lbound;

		if (neighb.x_[i] > ubound)
			neighb.x_[i] = ubound;
	}

	return neighb;
}


int main()
{
	sa::SA_Param<double> param;
	sa::SA_Result <double, Solution> res;
	std::vector<Solution> v_init_sol;
	Rastrigin_SA_OMP sa;

	double start_time, end_time;

	param.setGoal(sa::MIN);

	//param.setStopCond(sa::STEPS_LIM);
	//param.setStepsLim(1000);

	//param.setStopCond(sa::STALL_STEPS_LIM);
	//param.setStallStepsLim(100);

	param.setStopCond(sa::COST_LIM);
	param.setCostLim(0.1);

	//param.setStopCond(sa::TIME_LIM);
	//param.setTimeLim(30);

	//param.setStopCond(sa::TEMP_LIM);
	//param.setTempLim(0.1);

	param.setOfname("log.txt");

	param.setCommRate(100);
	
	param.setStopPolicy(sa::ONE);

	std::cout << param << "\n";
	std::cout << "Start SA...\n";

	start_time = get_wall_time();
	sa.run(param, v_init_sol, res);
	end_time = get_wall_time();

	std::cout << "End SA\n";
	std::cout << "Elapsed time: " << end_time - start_time << " second(s)\n\n";
	std::cout << res << "\n";

	_getch();

	return 0;
}