/**************************************************************************************************
Copyright (C) 2016 Andrey Tsyganov, UlSPU. All rights reserved.

Contact information
-------------------
Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
Web    :  http://lmm.ulspu.ru
e-mail :  andrew.tsyganov@gmail.com
**************************************************************************************************/


#ifndef GA_BASE_H
#define GA_BASE_H


//#define GA_BASE_DEBUG // Uncomment for debugging


#include <vector>
#include <string>
#include <iostream>
#include <ostream>
#include "../../contrib/timer/timer.h"
#include "../../contrib/randomc/randomc.h"


namespace ga
{


/**************************************************************************************************
GA types
**************************************************************************************************/


enum goal_t { MIN, MAX }; // Optimization goal 
enum cond_t { GEN_LIM, FITNESS_LIM, TIME_LIM }; // Stopping condition
enum stop_t { ALL, ONE }; // Threads stopping policy


typedef CRandomMersenne RndGen;


const size_t MAX_POP_SIZE = 50;


/**************************************************************************************************
Forward declarations
**************************************************************************************************/


template <typename TFitness>
class GA_Param;


template <typename TFitness>
std::ostream& operator<<(std::ostream& os, const GA_Param<TFitness>& param);


template <typename TFitness, typename TIndiv>
class GA_State;


/**************************************************************************************************
GA_Param: parameters class common for sequential and parallel versions of the GA
**************************************************************************************************/


template <typename TFitness>
class GA_Param
{
	template<typename F, typename S> friend class GA_BASE;
	template<typename F, typename S> friend class GA_SEQ;
	template<typename F, typename S> friend class GA_OMP;
	template<typename F, typename S> friend class GA_MPI;

public:
	// Constructor
	GA_Param();

	// Parameters setters
	void setGoal(goal_t goal);
	void setPopSize(size_t pop_size);
	void setEliteCount(size_t elite_count);
	void setDeadCount(size_t dead_count);
	void setTournamentCount(size_t tournament_count);
	void setMigrationCount(size_t migration_count);
	void setCrossoverFrac(double crossover_frac);
	void setStopCond(cond_t stop_cond);
	void setGenLim(size_t gen_lim);
	void setFitnessLim(TFitness fitness_lim);
	void setTimeLim(double time_lim);
	void setStopPolicy(stop_t stop_policy);
	void setCommRate(size_t comm_rate);
	void setOfname(std::string ofname);

	// Parameters getters
	goal_t getGoal() const;
	size_t getPopSize() const;
	size_t getEliteCount() const;
	size_t getDeadCount() const;
	size_t getTournamentCount() const;
	size_t getMigrationCount() const;
	double getCrossoverFrac() const;
	cond_t getStopCond() const;
	size_t getGenLim() const;
	TFitness getFitnessLim() const;
	double getTimeLim() const;
	stop_t getStopPolicy() const;
	size_t getCommRate() const;
	std::string getOfname() const;

	friend std::ostream& operator<< <>(std::ostream& os, const GA_Param<TFitness>& param);

protected:
	// Parameters
	goal_t goal_;
	size_t pop_size_;
	size_t elite_count_;
	size_t dead_count_;
	size_t tournament_count_;
	size_t migration_count_;
	double crossover_frac_;
	cond_t stop_cond_;
	size_t gen_lim_;
	TFitness fitness_lim_;
	double  time_lim_;
	stop_t stop_policy_;
	size_t comm_rate_;
	std::string ofname_;
};


template <typename TFitness>
GA_Param<TFitness>::GA_Param():
	goal_(MIN),
	pop_size_(20),
	elite_count_(2),
	dead_count_(2),
	tournament_count_(2),
	migration_count_(2),
	crossover_frac_(0.6),
	stop_cond_(GEN_LIM),
	gen_lim_(500),
	fitness_lim_(0),
	time_lim_(1),
	stop_policy_(ALL),
	comm_rate_(0)
{
}


template <typename TFitness>
void GA_Param<TFitness>::setGoal(goal_t goal)
{
	goal_ = goal;
}


template <typename TFitness>
void GA_Param<TFitness>::setPopSize(size_t pop_size)
{
	if ((pop_size_ > 0) && (pop_size <= MAX_POP_SIZE) && (pop_size >= elite_count_ + dead_count_))
	{
		pop_size_ = pop_size;
		return;
	}

	#ifdef GA_BASE_DEBUG
	std::cout << "Incorrect population size!\n";
	#endif
}


template <typename TFitness>
void GA_Param<TFitness>::setEliteCount(size_t elite_count)
{
	if (elite_count + dead_count_ <= pop_size_)
	{
		elite_count_ = elite_count;
		return;
	}

	#ifdef GA_BASE_DEBUG
	std::cout << "Incorrect elite counter!\n";
	#endif
}


template <typename TFitness>
void GA_Param<TFitness>::setDeadCount(size_t dead_count)
{
	if (dead_count + elite_count_ <= pop_size_)
	{
		dead_count_ = dead_count;
		return;
	}

	#ifdef GA_BASE_DEBUG
	std::cout << "Incorrect dead counter!\n";
	#endif
}


template <typename TFitness>
void GA_Param<TFitness>::setTournamentCount(size_t tournament_count)
{
	if (tournament_count > 0)
	{
		tournament_count_ = tournament_count;
		return;
	}

	#ifdef GA_BASE_DEBUG
	std::cout << "Incorrect tournament counter!\n";
	#endif
}


template <typename TFitness>
void GA_Param<TFitness>::setMigrationCount(size_t migration_count)
{
	if (migration_count > 0)
	{
		migration_count_ = migration_count;
		return;
	}

	#ifdef GA_BASE_DEBUG
	std::cout << "Incorrect migration counter!\n";
	#endif
}


template <typename TFitness>
void GA_Param<TFitness>::setCrossoverFrac(double crossover_frac)
{
	if ((crossover_frac >= 0) && (crossover_frac <= 1))
	{
		crossover_frac_ = crossover_frac;
		return;
	};

	#ifdef GA_BASE_DEBUG
	std::cout << "Incorrect crossover fraction!\n";
	#endif
}


template <typename TFitness>
void GA_Param<TFitness>::setStopCond(cond_t stop_cond)
{
	stop_cond_ = stop_cond;
}


template <typename TFitness>
void GA_Param<TFitness>::setGenLim(size_t gen_lim)
{
	gen_lim_ = gen_lim;
}


template <typename TFitness>
void GA_Param<TFitness>::setFitnessLim(TFitness fitness_lim)
{
	fitness_lim_ = fitness_lim;
}


template <typename TFitness>
void GA_Param<TFitness>::setTimeLim(double time_lim)
{
	if (time_lim >= 0)
	{
		time_lim_ = time_lim;
		return;
	};

	#ifdef GA_BASE_DEBUG
	std::cout << "Incorrect time limit!\n";
	#endif
}


template <typename TFitness>
void GA_Param<TFitness>::setStopPolicy(stop_t stop_policy)
{
	stop_policy_ = stop_policy;
}


template <typename TFitness>
void GA_Param<TFitness>::setCommRate(size_t comm_rate)
{
	comm_rate_ = comm_rate;
}


template <typename TFitness>
void GA_Param<TFitness>::setOfname(std::string ofname)
{
	ofname_ = ofname;
}


template <typename TFitness>
goal_t GA_Param<TFitness>::getGoal() const
{
	return goal_;
}


template <typename TFitness>
size_t GA_Param<TFitness>::getPopSize() const
{
	return pop_size_;
}


template <typename TFitness>
size_t GA_Param<TFitness>::getEliteCount() const
{
	return elite_count_;
}


template <typename TFitness>
size_t GA_Param<TFitness>::getDeadCount() const
{
	return dead_count_;
}


template <typename TFitness>
size_t GA_Param<TFitness>::getTournamentCount() const
{
	return tournament_count_;
}


template <typename TFitness>
size_t GA_Param<TFitness>::getMigrationCount() const
{
	return migration_count_;
}


template <typename TFitness>
double GA_Param<TFitness>::getCrossoverFrac() const
{
	return crossover_frac_;
}


template <typename TFitness>
cond_t GA_Param<TFitness>::getStopCond() const
{
	return stop_cond_;
}


template <typename TFitness>
size_t GA_Param<TFitness>::getGenLim() const
{
	return gen_lim_;
}


template <typename TFitness>
TFitness GA_Param<TFitness>::getFitnessLim() const
{
	return fitness_lim_;
}


template <typename TFitness>
double GA_Param<TFitness>::getTimeLim() const
{
	return dead_count_;
}


template <typename TFitness>
stop_t GA_Param<TFitness>::getStopPolicy() const
{
	return stop_policy_;
}


template <typename TFitness>
size_t GA_Param<TFitness>::getCommRate() const
{
	return comm_rate_;
}


template <typename TFitness>
std::string GA_Param<TFitness>::getOfname() const
{
	return ofname_;
}


template <typename TFitness>
std::ostream& operator<<(std::ostream& os, const GA_Param<TFitness>& param)
{
	const char goal_names[2][5] = { "MIN", "MAX" };
	const char cond_names[3][20] = { "GEN_LIM", "FITNESS_LIM", "TIME_LIM" };
	const char stop_names[2][5] = { "ALL", "ONE" };

	os << "goal_ = " << goal_names[param.goal_] << "\n";
	os << "pop_size_ = " << param.pop_size_ << "\n";
	os << "elite_count_ = " << param.elite_count_ << "\n";
	os << "dead_count_ = " << param.dead_count_ << "\n";
	os << "tournament_count_ = " << param.tournament_count_ << "\n";
	os << "migration_count_ = " << param.migration_count_ << "\n";
	os << "crossover_frac_ = " << param.crossover_frac_ << "\n";
	os << "stop_cond_ = " << cond_names[param.stop_cond_] << "\n";
	os << "gen_lim_ = " << param.gen_lim_ << "\n";
	os << "fitness_lim_ = " << param.fitness_lim_ << "\n";
	os << "time_lim_ = " << param.time_lim_ << "\n";
	os << "stop_policy_ = " << stop_names[param.stop_policy_] << "\n";
	os << "comm_rate_ = " << param.comm_rate_ << "\n";
	os << "ofname_ = " << param.ofname_ << "\n";

	return os;
}


/**************************************************************************************************
GA_State: class for storing current state of the GA
**************************************************************************************************/


template <typename TFitness, typename TIndiv>
class GA_State
{
public:
	GA_State();

	int rank_;
	int pnum_;
	double start_time_;
	size_t gen_;
	size_t crossover_count_;
	size_t mutation_count_;
	size_t alive_count_;
	std::vector<TIndiv> population_;
	std::vector<TFitness> fitnesses_;
	std::vector<TIndiv> new_population_;

	friend std::ostream& operator<< <>(std::ostream& os, const GA_State<TFitness, TIndiv>& state); // Stream output
	void sortPopulation(goal_t goal);
};


template <typename TFitness, typename TIndiv>
GA_State<TFitness, TIndiv>::GA_State():
	rank_(0),
	pnum_(1),
	start_time_(0),
	gen_(0),
	crossover_count_(0),
	mutation_count_(0),
	alive_count_(0)
{
}


template <typename TFitness, typename TIndiv>
std::ostream& operator<<(std::ostream& os, const GA_State<TFitness, TIndiv>& state)
{
	os << "rank_ = " << state.rank_ << "\n";
	os << "pnum_ = " << state.pnum_ << "\n";
	os << "start_time_ = " << state.start_time_ << "\n";
	os << "gen_ = " << state.gen_ << "\n";
	os << "crossover_count_ = " << state.crossover_count_ << "\n";
	os << "mutation_count_ = " << state.mutation_count_ << "\n";
	os << "alive_count_ = " << state.alive_count_ << "\n";

	os << "population_:\n";
	for (size_t i = 0; i < state.population_.size(); i++)
		os << state.population_[i] << "\n";

	os << "fitnesses_:\n";
	for (size_t i = 0; i < state.fitnesses_.size(); i++)
		os << state.fitnesses_[i] << "\n";

	os << "new_population_:\n";
	for (size_t i = 0; i < state.new_population_.size(); i++)
		os << state.new_population_[i] << "\n";
	return os;
}


template <typename TFitness, typename TIndiv>
void GA_State<TFitness, TIndiv>::sortPopulation(goal_t goal)
{
	for (size_t i = 0; i < population_.size() - 1; i++)
		for (size_t j = 0; j < population_.size() - (i + 1); j++)
			if (((goal == MIN) && (fitnesses_[j] > fitnesses_[j + 1])) || ((goal == MAX) && (fitnesses_[j] < fitnesses_[j + 1])))
			{
				TFitness temp_fit = fitnesses_[j];
				TIndiv temp_indiv = population_[j];

				fitnesses_[j] = fitnesses_[j + 1];
				population_[j] = population_[j + 1];

				fitnesses_[j + 1] = temp_fit;
				population_[j + 1] = temp_indiv;
			}
}


/**************************************************************************************************
GA_Result: class for storing results of the GA
**************************************************************************************************/


template <typename TFitness, typename TIndiv>
class GA_Result
{
public:
	std::vector< GA_State<TFitness, TIndiv> > v_state_;

	friend std::ostream& operator<< <>(std::ostream& os, const GA_Result<TFitness, TIndiv>& result); // Stream output
};


template <typename TFitness, typename TIndiv>
std::ostream& operator<<(std::ostream& os, const GA_Result<TFitness, TIndiv>& result)
{
	for (size_t i = 0; i < result.v_state_.size(); i++)
	{
		os << "Thread[" << i << "]:\n";
		os << result.v_state_[i] << "\n";
	}

	return os;
}


/**************************************************************************************************
GA_BASE: base abstract class from which abstract sequential and parallel algorithms classes
are derived
**************************************************************************************************/


template <typename TFitness, typename TIndiv>
class GA_BASE
{
public:
	// These methods may be overriden in derived user classes
	virtual int seed(int rank, int pnum); // Random seed
	virtual void output(const GA_Param<TFitness>& param, const GA_State<TFitness, TIndiv>& state, std::ostream& os); // Output current state

	// This method is implemented in derived classes GA_BASE, GA_OMP and GA_MPI
	virtual void run(const GA_Param<TFitness>& param, const std::vector<TIndiv>& initial_pop, GA_Result<TFitness, TIndiv>& res) = 0; // Main method

	virtual TFitness fitness(const TIndiv& indiv, const GA_State<TFitness, TIndiv>& state) = 0; // Evaluate fitness
	virtual TIndiv generate(const GA_State<TFitness, TIndiv>& state, RndGen& rnd) = 0; // Generate individual
	virtual TIndiv crossover(const TIndiv& parent1, const TIndiv& parent2, const GA_State<TFitness, TIndiv>& state, RndGen& rnd) = 0; // Perform crossover
	virtual TIndiv mutation(const TIndiv& indiv, const GA_State<TFitness, TIndiv>& state, RndGen& rnd) = 0; // Perform mutation 

protected:
	void setInitialPopulation(const std::vector<TIndiv>& initial_pop, GA_State<TFitness, TIndiv>& state, RndGen& rnd); // Set initial population
	boolean stopCondition(const GA_Param<TFitness>& param, const GA_State<TFitness, TIndiv>& state); // Stop condition checker
};


template <typename TFitness, typename TIndiv>
int GA_BASE<TFitness, TIndiv>::seed(int rank, int pnum)
{
	return (rank << 4) ^ pnum;
}


template <typename TFitness, typename TIndiv>
void GA_BASE<TFitness, TIndiv>::output(const GA_Param<TFitness>& param, const GA_State<TFitness, TIndiv>& state, std::ostream& os)
{
}


template <typename TFitness, typename TIndiv>
void GA_BASE<TFitness, TIndiv>::setInitialPopulation(const std::vector<TIndiv>& initial_pop, GA_State<TFitness, TIndiv>& state, RndGen& rnd)
{
	if (initial_pop.size() == 0)
	{
		for (size_t i = 0; i < state.population_.size(); i++)
		{
			state.population_[i] = generate(state, rnd);
			state.fitnesses_[i] = fitness(state.population_[i], state);
		}
	}
	else
	{
		if (initial_pop.size() >= state.population_.size())
		{
			for (size_t i = 0; i < state.population_.size(); i++)
			{
				state.population_[i] = initial_pop[i];
				state.fitnesses_[i] = fitness(state.population_[i], state);
			}
		}
		else
		{
			for (size_t i = 0; i < state.population_.size(); i++)
			{
				state.population_[i] = initial_pop[i % initial_pop.size()];
				state.fitnesses_[i] = fitness(state.population_[i], state);
			}
		}
	}
	
}


template <typename TFitness, typename TIndiv>
boolean GA_BASE<TFitness, TIndiv>::stopCondition(const GA_Param<TFitness>& param, const GA_State<TFitness, TIndiv>& state)
{
	switch (param.stop_cond_)
	{
		case GEN_LIM:
			return (state.gen_ >= param.gen_lim_ ? true : false);

		case FITNESS_LIM:
			if (param.goal_ == MIN)
				return (state.fitnesses_[0] <= param.fitness_lim_ ? true : false);
			else
				return (state.fitnesses_[0] >= param.fitness_lim_ ? true : false);

		case TIME_LIM:
			return ((get_wall_time() - state.start_time_) >= param.time_lim_ ? true : false);
	}
}


}


#endif
