/**************************************************************************************************
Copyright (C) 2016 Andrey Tsyganov, UlSPU. All rights reserved.

Contact information
-------------------
Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
Web    :  http://lmm.ulspu.ru
e-mail :  andrew.tsyganov@gmail.com
**************************************************************************************************/


#ifndef GA_OMP_H
#define GA_OMP_H


//#define GA_OMP_DEBUG // Uncomment for debugging


#include <omp.h>
#include <fstream>
#include "ga_base.h"
#include "../../inc/utils/utils.h"


namespace ga
{


/**************************************************************************************************
GA_OMP: abstract class from which user OpenMP algorithm class is derived
**************************************************************************************************/


template <typename TFitness, typename TIndiv>
class GA_OMP: public GA_BASE < TFitness, TIndiv >
{
public:
	virtual void run(const GA_Param<TFitness>& param, const std::vector<TIndiv>& v_init_sol, GA_Result<TFitness, TIndiv>& res); // Main method
};


template <class TFitness, class TIndiv>
void GA_OMP<TFitness, TIndiv>::run(const GA_Param<TFitness>& param, const std::vector<TIndiv>& initial_pop, GA_Result<TFitness, TIndiv>& res)
{
	int nthreads = omp_get_max_threads();
	std::vector<omp_lock_t> population_locks;
	bool stop_shared = false;

	res.v_state_.resize(nthreads);
	for (int i = 0; i < nthreads; i++)
	{
		res.v_state_[i].rank_ = i;
		res.v_state_[i].pnum_ = nthreads;
	}

	if (param.comm_rate_ > 0)
	{
		population_locks.resize(nthreads);
		for (int i = 0; i < nthreads; i++)
			omp_init_lock(&population_locks[i]);
	}

	#pragma omp parallel num_threads(nthreads)
	{
		int rank = omp_get_thread_num();
		GA_State<TFitness, TIndiv>& state = res.v_state_[rank];

		// Begin: obtain ostream

		std::ofstream of;
		std::streambuf* buf = OSBuf(param.ofname_, of, rank, nthreads);
		std::ostream os(buf);

		// End: obtain ostream

		state.start_time_ = get_wall_time();

		int left = (rank - 1 < 0 ? nthreads - 1 : rank - 1); // Determine rank of the left neighbour

		RndGen rnd(seed(rank, nthreads));

		state.population_.resize(param.pop_size_);
		state.fitnesses_.resize(param.pop_size_);
		state.new_population_.resize(param.pop_size_ - param.elite_count_);

		state.crossover_count_ = (size_t)((param.pop_size_ - param.elite_count_) * param.crossover_frac_);
		state.mutation_count_ = param.pop_size_ - param.elite_count_ - state.crossover_count_;
		state.alive_count_ = param.pop_size_ - param.dead_count_;

		state.gen_ = 0;
		setInitialPopulation(initial_pop, state, rnd);
		state.sortPopulation(param.goal_); // Dead individuals are now at the end

		#pragma omp barrier // Let all threads set initial population

		#ifdef GA_OMP_DEBUG
		#pragma omp critical
		std::cout << state << "\n";
		#endif

		output(param, state, os);

		bool stop_priv = stopCondition(param, state);

		if ((param.stop_policy_ == ONE) && stop_priv)
		{
			stop_shared = true;
			#pragma omp flush(stop_shared)
		}

		// Begin: main loop
		while (!(param.stop_policy_ == ONE ? stop_shared : stop_priv))
		{
			state.gen_++;

			// Apply crossover on alive individuals to produce crossover children
			for (size_t i = 0; i < state.crossover_count_; i++)
			{
				// Tournamnet selection of the first parent
				size_t ind1 = rnd.IRandom(0, state.alive_count_ - 1);
				for (size_t k = 1; k < param.tournament_count_; k++)
				{
					size_t ind = rnd.IRandom(0, state.alive_count_ - 1);
					if (ind < ind1)
						ind1 = ind;
				}

				// Tournamnet selection of the second parent
				size_t ind2 = rnd.IRandom(0, state.alive_count_ - 1);
				for (size_t k = 1; k < param.tournament_count_; k++)
				{
					size_t ind = rnd.IRandom(0, state.alive_count_ - 1);
					if (ind < ind2)
						ind2 = ind;
				}

				// Apply crossover to the selected parents to produce children (parents may coincide, best parent goes first)
				if (state.fitnesses_[ind1] < state.fitnesses_[ind2])
					state.new_population_[i] = crossover(state.population_[ind1], state.population_[ind2], state, rnd);
				else
					state.new_population_[i] = crossover(state.population_[ind2], state.population_[ind1], state, rnd);
			}

			// Apply mutation on alive individuals to produce mutation children
			for (size_t i = 0; i < state.mutation_count_; i++)
				state.new_population_[state.crossover_count_ + i] = mutation(state.population_[rnd.IRandom(0, state.alive_count_ - 1)], state, rnd);

			if ((param.comm_rate_ > 0) && (state.gen_ % param.comm_rate_ == 0))
			{
				// Begin: communication

				omp_set_lock(&population_locks[left]);

				//#pragma omp critical
				//std::cout << "thread " << rank << " starts communication\n";

				// Replace random members of the new population with migrants (best members of the left neighbour)
				// Some migrants "die" when IRandom() yields the same indices
				for (size_t i = 0; i < param.migration_count_; i++)
				{
					size_t ind = rnd.IRandom(0, state.new_population_.size() - 1);
					state.new_population_[ind] = res.v_state_[left].population_[i];
				}

				omp_unset_lock(&population_locks[left]);

				//#pragma omp critical
				//std::cout << "thread " << rank << " finishes communication\n";

				// End: communication
			}

			if (param.comm_rate_ > 0)
				omp_set_lock(&population_locks[rank]);

			//#pragma omp critical
			//std::cout << "thread " << rank << " starts population update\n";

			// Update population (elite individuals are already at the beginning)
			for (size_t i = param.elite_count_; i < param.pop_size_; i++)
			{
				state.population_[i] = state.new_population_[i - param.elite_count_];
				state.fitnesses_[i] = fitness(state.population_[i], state);
			}

			state.sortPopulation(param.goal_);

			if (param.comm_rate_ > 0)
				omp_unset_lock(&population_locks[rank]);

			//#pragma omp critical
			//std::cout << "thread " << rank << " finishes population update\n";

			#ifdef GA_OMP_DEBUG
			#pragma omp critical
			std::cout << state << "\n";
			#endif

			output(param, state, os);

			stop_priv = stopCondition(param, state);

			if ((param.stop_policy_ == ONE) && stop_priv)
			{
				stop_shared = true;
				#pragma omp flush(stop_shared)
			}
		}

		// End: main loop
	}

	if (param.comm_rate_ > 0)
	{
		for (int i = 0; i < nthreads; i++)
			omp_destroy_lock(&population_locks[i]);
	}
}


}


#endif
