/**************************************************************************************************
Copyright (C) 2016 Andrey Tsyganov, UlSPU. All rights reserved.

Contact information
-------------------
Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
Web    :  http://lmm.ulspu.ru
e-mail :  andrew.tsyganov@gmail.com
**************************************************************************************************/


#ifndef GA_SEQ_H
#define GA_SEQ_H


//#define GA_SEQ_DEBUG // Uncomment for debugging


#include <fstream>
#include "ga_base.h"
#include "../../inc/utils/utils.h"


namespace ga
{


/**************************************************************************************************
GA_SEQ: base abstract class from which user sequential algorithm classes are derived
**************************************************************************************************/


template <typename TFitness, typename TIndiv>
class GA_SEQ: public GA_BASE < TFitness, TIndiv >
{
public:
	virtual void run(const GA_Param<TFitness>& param, const std::vector<TIndiv>& initial_pop, GA_Result<TFitness, TIndiv>& res); // Main method
};


template <class TFitness, class TIndiv>
void GA_SEQ<TFitness, TIndiv>::run(const GA_Param<TFitness>& param, const std::vector<TIndiv>& initial_pop, GA_Result<TFitness, TIndiv>& res)
{
	GA_State<TFitness, TIndiv> state;

	// Begin: obtain ostream

	std::ofstream of;
	std::streambuf* buf = OSBuf(param.ofname_, of, state.rank_, state.pnum_);
	std::ostream os(buf);

	// End: obtain ostream

	state.start_time_ = get_wall_time();

	RndGen rnd(seed(state.rank_, state.pnum_));

	state.population_.resize(param.pop_size_);
	state.fitnesses_.resize(param.pop_size_);
	state.new_population_.resize(param.pop_size_ - param.elite_count_);

	state.crossover_count_ = (size_t)((param.pop_size_ - param.elite_count_) * param.crossover_frac_);
	state.mutation_count_ = param.pop_size_ - param.elite_count_ - state.crossover_count_;
	state.alive_count_ = param.pop_size_ - param.dead_count_;

	state.gen_ = 0;
	setInitialPopulation(initial_pop, state, rnd);
	state.sortPopulation(param.goal_); // Dead individuals are now at the end

	#ifdef GA_SEQ_DEBUG
	std::cout << state << "\n";
	#endif

	output(param, state, os);

	// Begin: main loop
	while (!stopCondition(param, state))
	{
		state.gen_++;

		// Apply crossover on alive individuals to produce crossover children
		for (size_t i = 0; i < state.crossover_count_; i++)
		{
			// Tournamnet selection of the first parent
			size_t ind1 = rnd.IRandom(0, state.alive_count_ - 1);
			for (size_t k = 1; k < param.tournament_count_; k++)
			{
				size_t ind = rnd.IRandom(0, state.alive_count_ - 1);
				if (ind < ind1)
					ind1 = ind;
			}

			// Tournamnet selection of the second parent
			size_t ind2 = rnd.IRandom(0, state.alive_count_ - 1);
			for (size_t k = 1; k < param.tournament_count_; k++)
			{
				size_t ind = rnd.IRandom(0, state.alive_count_ - 1);
				if (ind < ind2)
					ind2 = ind;
			}

			// Apply crossover to the selected parents to produce children (parents may coincide, best parent goes first)
			if (state.fitnesses_[ind1] < state.fitnesses_[ind2])
				state.new_population_[i] = crossover(state.population_[ind1], state.population_[ind2], state, rnd);
			else
				state.new_population_[i] = crossover(state.population_[ind2], state.population_[ind1], state, rnd);
		}

		// Apply mutation on alive individuals to produce mutation children
		for (size_t i = 0; i < state.mutation_count_; i++)
			state.new_population_[state.crossover_count_ + i] = mutation(state.population_[rnd.IRandom(0, state.alive_count_ - 1)], state, rnd);

		// Update population (elite individuals are already at the beginning)
		for (size_t i = param.elite_count_; i < param.pop_size_; i++)
		{
			state.population_[i] = state.new_population_[i - param.elite_count_];
			state.fitnesses_[i] = fitness(state.population_[i], state);
		}

		state.sortPopulation(param.goal_);

		#ifdef GA_SEQ_DEBUG
		std::cout << state << "\n";
		#endif

		output(param, state, os);
	}

	// End: main loop

	res.v_state_.push_back(state);
}


}


#endif
