/**************************************************************************************************
Copyright (C) 2016 Andrey Tsyganov, UlSPU. All rights reserved.

Contact information
-------------------
Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
Web    :  http://lmm.ulspu.ru
e-mail :  andrew.tsyganov@gmail.com
**************************************************************************************************/


#ifndef SA_BASE_H
#define SA_BASE_H


#define SA_BASE_DEBUG // Uncomment for debugging


#include <vector>
#include <string>
#include <iostream>
#include <ostream>
#include "../../contrib/timer/timer.h"
#include "../../contrib/randomc/randomc.h"


namespace sa
{


/**************************************************************************************************
SA types
**************************************************************************************************/


enum goal_t { MIN, MAX }; // Optimization goal 
enum cond_t { STEPS_LIM, STALL_STEPS_LIM, COST_LIM, TIME_LIM, TEMP_LIM }; // Stopping condition
enum cool_t { EXP, LINEAR, FAST, BOLTZ }; // Cooling schedule
enum stop_t { ALL, ONE }; // Threads stopping policy


typedef CRandomMersenne RndGen;


/**************************************************************************************************
Forward declarations
**************************************************************************************************/


template <typename TCost>
class SA_Param;


template <typename TCost>
std::ostream& operator<<(std::ostream& os, const SA_Param<TCost>& param);


template <typename TCost, typename TSolution>
class SA_State;


/**************************************************************************************************
SA_Param: parameters class common for sequential and parallel versions of the SA algorithm
**************************************************************************************************/


template <typename TCost>
class SA_Param
{
	template<typename C, typename S> friend class SA_BASE;
	template<typename C, typename S> friend class SA_SEQ;
	template<typename C, typename S> friend class SA_OMP;
	template<typename C, typename S> friend class SA_MPI;

public:
	// Constructor
	SA_Param();

	// Parameters setters
	void setGoal(goal_t goal);
	void setStopCond(cond_t stop_cond);
	void setStepsLim(size_t steps_lim);
	void setStallStepsLim(size_t steps_lim);
	void setCostLim(TCost cost_lim);
	void setTimeLim(double time_lim);
	void setTempLim(double temp_lim);
	void setStepLen(size_t step_len);
	void setInitialTemp(double initial_temp);
	void setInitialProb(double initial_prob);
	void setCooling(cool_t cooling);
	void setStopPolicy(stop_t stop_policy);
	void setCoolingRate(double cooling_rate);
	void setCommRate(size_t comm_rate);
	void setOfname(std::string ofname);

	// Parameters getters
	goal_t getGoal() const;
	cond_t getStopCond() const;
	size_t getStepsLim() const;
	size_t getStallStepsLim() const;
	TCost  getCostLim() const;
	double getTimeLim() const;
	double getTempLim() const;
	size_t getStepLen() const;
	double getInitialTemp() const;
	double getInitialProb() const;
	cool_t getCooling() const;
	stop_t getStopPolicy() const;
	double getCoolingRate() const;
	size_t getCommRate() const;
	std::string getOfname() const;

	friend std::ostream& operator<< <>(std::ostream& os, const SA_Param<TCost>& param);

protected:
	// Parameters
	goal_t goal_;
	cond_t stop_cond_;
	size_t steps_lim_;
	size_t stall_steps_lim_;
	TCost  cost_lim_;
	double time_lim_;
	double temp_lim_;
	size_t step_len_;
	bool   auto_heat_;
	double initial_temp_;
	double initial_prob_;
	cool_t cooling_;
	double cooling_rate_;
	stop_t stop_policy_;
	size_t comm_rate_;
	std::string ofname_;
};


template <typename TCost>
SA_Param<TCost>::SA_Param():
	goal_(MIN),
	stop_cond_(STEPS_LIM),
	steps_lim_(500),
	stall_steps_lim_(100),
	cost_lim_(0),
	time_lim_(1),
	temp_lim_(0.01),
	step_len_(50),
	auto_heat_(true),
	initial_temp_(1.0),
	initial_prob_(0.9),
	cooling_(EXP),
	cooling_rate_(0.95),
	stop_policy_(ALL),
	comm_rate_(0)
{
}


template <typename TCost>
void SA_Param<TCost>::setGoal(goal_t goal)
{
	goal_ = goal;
}


template <typename TCost>
void SA_Param<TCost>::setStopCond(cond_t stop_cond)
{
	stop_cond_ = stop_cond;
}


template <typename TCost>
void SA_Param<TCost>::setStepsLim(size_t steps_lim)
{
	steps_lim_ = steps_lim;
}


template <typename TCost>
void SA_Param<TCost>::setStallStepsLim(size_t stall_steps_lim)
{
	stall_steps_lim_ = stall_steps_lim;
}


template <typename TCost>
void SA_Param<TCost>::setCostLim(TCost cost_lim)
{
	cost_lim_ = cost_lim;
}


template <typename TCost>
void SA_Param<TCost>::setTimeLim(double time_lim)
{
	if (time_lim >= 0)
	{
		time_lim_ = time_lim;
		return;
	};

	#ifdef SA_BASE_DEBUG
	std::cout << "Incorrect time limit!\n";
	#endif
}


template <typename TCost>
void SA_Param<TCost>::setTempLim(double temp_lim)
{
	if (temp_lim > 0)
	{
		temp_lim_ = temp_lim;
		return;
	};

	#ifdef SA_BASE_DEBUG
	std::cout << "Incorrect temperature limit!\n";
	#endif
}


template <typename TCost>
void SA_Param<TCost>::setStepLen(size_t step_len)
{
	if (step_len > 0)
	{
		step_len_ = step_len;
		return;
	};

	#ifdef SA_BASE_DEBUG
	std::cout << "Incorrect step length!\n";
	#endif
}


template <typename TCost>
void SA_Param<TCost>::setInitialTemp(double initial_temp)
{
	if (initial_temp > 0)
	{
		initial_temp_ = initial_temp;
		return;
	};

	#ifdef SA_BASE_DEBUG
	std::cout << "Incorrect initial temperature!\n";
	#endif
}


template <typename TCost>
void SA_Param<TCost>::setInitialProb(double initial_prob)
{
	if ((initial_prob > 0) && (initial_prob < 1))
	{
		initial_prob_ = initial_prob;
		return;
	};

	#ifdef SA_BASE_DEBUG
	std::cout << "Incorrect initial probability!\n";
	#endif
}


template <typename TCost>
void SA_Param<TCost>::setCooling(cool_t cooling)
{
	cooling_ = cooling;
}


template <typename TCost>
void SA_Param<TCost>::setStopPolicy(stop_t stop_policy)
{
	stop_policy_ = stop_policy;
}


template <typename TCost>
void SA_Param<TCost>::setCoolingRate(double cooling_rate)
{
	if (cooling_rate > 0)
	{
		cooling_rate_ = cooling_rate;
		return;
	};

	#ifdef SA_BASE_DEBUG
	std::cout << "Incorrect cooling rate!\n";
	#endif
}


template <typename TCost>
void SA_Param<TCost>::setCommRate(size_t comm_rate)
{
	comm_rate_ = comm_rate;
}


template <typename TCost>
void SA_Param<TCost>::setOfname(std::string ofname)
{
	ofname_ = ofname;
}


template <typename TCost>
goal_t SA_Param<TCost>::getGoal() const
{
	return goal_;
}


template <typename TCost>
cond_t SA_Param<TCost>::getStopCond() const
{
	return stop_cond_;
}


template <typename TCost>
size_t SA_Param<TCost>::getStepsLim() const
{
	return steps_lim_;
}


template <typename TCost>
size_t SA_Param<TCost>::getStallStepsLim() const
{
	return stall_steps_lim_;
}


template <typename TCost>
TCost SA_Param<TCost>::getCostLim() const
{
	return cost_lim_;
}


template <typename TCost>
double SA_Param<TCost>::getTimeLim() const
{
	return time_lim_;
}


template <typename TCost>
double SA_Param<TCost>::getTempLim() const
{
	return temp_lim_;
}


template <typename TCost>
size_t SA_Param<TCost>::getStepLen() const
{
	return step_len_;
}


template <typename TCost>
double SA_Param<TCost>::getInitialTemp() const
{
	return initial_temp_;
}


template <typename TCost>
double SA_Param<TCost>::getInitialProb() const
{
	return initial_prob_;
}


template <typename TCost>
cool_t SA_Param<TCost>::getCooling() const
{
	return cooling_;
}


template <typename TCost>
stop_t SA_Param<TCost>::getStopPolicy() const
{
	return stop_policy_;
}


template <typename TCost>
double SA_Param<TCost>::getCoolingRate() const
{
	return cooling_rate_;
}


template <typename TCost>
size_t SA_Param<TCost>::getCommRate() const
{
	return comm_rate_;
}


template <typename TCost>
std::string SA_Param<TCost>::getOfname() const
{
	return ofname_;
}


template <typename TCost>
std::ostream& operator<<(std::ostream& os, const SA_Param<TCost>& param)
{
	const char goal_names[2][5] = { "MIN", "MAX" };
	const char cond_names[5][20] = { "STEPS_LIM", "STALL_STEPS_LIM", "COST_LIM", "TIME_LIM", "TEMP_LIM" };
	const char cool_names[4][10] = { "EXP", "LINEAR", "FAST", "BOLTZ" };
	const char stop_names[2][5] = { "ALL", "ONE" };

	os << "goal_ = " << goal_names[param.goal_] << "\n";
	os << "stop_cond_ = " << cond_names[param.stop_cond_] << "\n";
	os << "steps_lim_ = " << param.steps_lim_ << "\n";
	os << "stall_steps_lim_ = " << param.stall_steps_lim_ << "\n";
	os << "cost_lim_ = " << param.cost_lim_ << "\n";
	os << "time_lim_ = " << param.time_lim_ << "\n";
	os << "temp_lim_ = " << param.temp_lim_ << "\n";
	os << "step_len_ = " << param.step_len_ << "\n";
	os << "auto_heat_ = " << std::boolalpha << param.auto_heat_ << "\n";
	os << "initial_temp_ = " << param.initial_temp_ << "\n";
	os << "initial_prob_ = " << param.initial_prob_ << "\n";
	os << "cooling_ = " << cool_names[param.cooling_] << "\n";
	os << "cooling_rate_ = " << param.cooling_rate_ << "\n";
	os << "stop_policy_ = " << stop_names[param.stop_policy_] << "\n";
	os << "comm_rate_ = " << param.comm_rate_ << "\n";
	os << "ofname_ = " << param.ofname_ << "\n";

	return os;
}


/**************************************************************************************************
SA_State: class for storing current state of the SA algorithm
**************************************************************************************************/


template <typename TCost, typename TSolution>
class SA_State
{
public:
	SA_State();

	int rank_;
	int pnum_;
	double start_time_;
	double initial_temp_;
	double temp_;
	size_t step_;
	size_t step_iter_;
	size_t stall_step_;
	bool no_change_;
	TSolution current_;
	TSolution new_;
	TSolution best_;
	TCost current_cost_;
	TCost new_cost_;
	TCost best_cost_;

	friend std::ostream& operator<< <>(std::ostream& os, const SA_State<TCost, TSolution>& state); // Stream output
};


template <typename TCost, typename TSolution>
SA_State<TCost, TSolution>::SA_State():
	rank_(0),
	pnum_(1),
	start_time_(0),
	initial_temp_(0),
	temp_(0),
	step_(0),
	step_iter_(0),
	stall_step_(0),
	no_change_(false)
{
}


template <typename TCost, typename TSolution>
std::ostream& operator<<(std::ostream& os, const SA_State<TCost, TSolution>& state)
{
	os << "rank_ = " << state.rank_ << "\n";
	os << "pnum_ = " << state.pnum_ << "\n";
	os << "start_time_ = " << state.start_time_ << "\n";
	os << "initial_temp_ = " << state.initial_temp_ << "\n";
	os << "temp_ = " << state.temp_ << "\n";
	os << "step_ = " << state.step_ << "\n";
	os << "step_iter_ = " << state.step_iter_ << "\n";
	os << "stall_step_ = " << state.stall_step_ << "\n";
	os << "no_change_ = " << std::boolalpha << state.no_change_ << "\n";
	os << "current_:\n";
	os << state.current_ << "\n";
	os << "current_cost_ = " << state.current_cost_ << "\n";
	os << "new_:\n";
	os << state.new_ << "\n";
	os << "new_cost_ = " << state.new_cost_ << "\n";
	os << "best_:\n";
	os << state.best_ << "\n";
	os << "best_cost_ = " << state.best_cost_ << "\n";

	return os;
}


/**************************************************************************************************
SA_Result: class for storing results of the SA algorithm
**************************************************************************************************/


template <typename TCost, typename TSolution>
class SA_Result
{
public:
	std::vector< SA_State<TCost, TSolution> > v_state_;

	friend std::ostream& operator<< <>(std::ostream& os, const SA_Result<TCost, TSolution>& result); // Stream output
};


template <typename TCost, typename TSolution>
std::ostream& operator<<(std::ostream& os, const SA_Result<TCost, TSolution>& result)
{
	for (size_t i = 0; i < result.v_state_.size(); i++)
	{
		os << "Thread[" << i << "]:\n";
		os << result.v_state_[i] << "\n";
	}

	return os;
}


/**************************************************************************************************
SA_BASE: base abstract class from which abstract sequential and parallel algorithms classes
are derived
**************************************************************************************************/


template <typename TCost, typename TSolution>
class SA_BASE
{
public:
	// These methods may be overriden in derived user classes
	virtual int seed(int rank, int pnum); // Random seed
	virtual void output(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state, std::ostream& os); // Output current state

	// This method is implemented in derived classes SA_SEQ, SA_OMP and SA_MPI
	virtual void run(const SA_Param<TCost>& param, const std::vector<TSolution>& v_init_sol, SA_Result<TCost, TSolution>& res) = 0; // Main method

	// These pure virtual methods are implemented in problem-specific user classes
	virtual TCost cost(const TSolution& solution, const SA_State<TCost, TSolution>& state) = 0;  // Evaluate cost
	virtual TSolution generate(const SA_State<TCost, TSolution>& state, RndGen& rnd) = 0; // Generate solution
	virtual TSolution neighbour(const TSolution& solution, const SA_State<TCost, TSolution>& state, RndGen& rnd) = 0; // Generate neighbour

protected:
	bool stopCondition(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state); // Stop condition checker
	bool acceptNew(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state, RndGen& rnd); // Acceptance function
	bool currentBest(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state); // Check if current is the best solution
	double initialTemp(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state, RndGen& rnd); // Calculate initial temperature
	double newTemp(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state); // Cooling function
};


template <typename TCost, typename TSolution>
int SA_BASE<TCost, TSolution>::seed(int rank, int pnum)
{
	return (rank << 4) ^ pnum;
}


template <typename TCost, typename TSolution>
void SA_BASE<TCost, TSolution>::output(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state, std::ostream& os)
{
}


template <typename TCost, typename TSolution>
bool SA_BASE<TCost, TSolution>::acceptNew(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state, RndGen& rnd)
{
	if (param.goal_ == MIN)
		return (state.new_cost_ < state.current_cost_ ? true : exp(-(state.new_cost_ - state.current_cost_) / state.temp_) > rnd.Random());
	else
		return (state.new_cost_ > state.current_cost_ ? true : exp((state.new_cost_ - state.current_cost_) / state.temp_) > rnd.Random());
}


template <typename TCost, typename TSolution>
bool SA_BASE<TCost, TSolution>::currentBest(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state)
{
	return ((param.goal_ == MIN) && (state.current_cost_ < state.best_cost_) || (param.goal_ == MAX) && (state.current_cost_ > state.best_cost_));
}


template <typename TCost, typename TSolution>
double SA_BASE<TCost, TSolution>::initialTemp(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state, RndGen& rnd)
{
	double initial_temp;

	if (param.auto_heat_)
	{
		double max_delta = 0;
		for (int i = 0; i < 50; i++)
		{
			double delta = fabs((double)(cost(generate(state, rnd), state) - cost(generate(state, rnd), state)));
			if (delta > max_delta)
				max_delta = delta;
		}

		initial_temp = -max_delta / log(param.initial_prob_);
		initial_temp = max(initial_temp, param.initial_temp_);
	}
	else
	{
		initial_temp = param.initial_temp_;
	}

	return initial_temp;
}


template <typename TCost, typename TSolution>
double SA_BASE<TCost, TSolution>::newTemp(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state)
{
	switch (param.cooling_)
	{
		case EXP:
			return state.temp_ * param.cooling_rate_;

		case LINEAR:
			return state.temp_ - param.cooling_rate_ * state.step_;

		case FAST:
			return state.initial_temp_ / (1 + state.step_);

		case BOLTZ:
			return state.initial_temp_ / log(1 + state.step_);
	}
}


template <typename TCost, typename TSolution>
bool SA_BASE<TCost, TSolution>::stopCondition(const SA_Param<TCost>& param, const SA_State<TCost, TSolution>& state)
{
	switch (param.stop_cond_)
	{
		case STEPS_LIM:
			return (state.step_ >= param.steps_lim_ ? true : false);

		case STALL_STEPS_LIM:
			return (state.stall_step_ >= param.stall_steps_lim_ ? true : false);

		case COST_LIM:
			if (param.goal_ == MIN)
				return (state.best_cost_ <= param.cost_lim_ ? true : false);
			else
				return (state.best_cost_ >= param.cost_lim_ ? true : false);

		case TIME_LIM:
			return ((get_wall_time() - state.start_time_) >= param.time_lim_ ? true : false);

		case TEMP_LIM:
			return (state.temp_ <= param.temp_lim_ ? true : false);
	}
}


}


#endif
