/**************************************************************************************************
Copyright (C) 2016 Andrey Tsyganov, UlSPU. All rights reserved.

Contact information
-------------------
Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
Web    :  http://lmm.ulspu.ru
e-mail :  andrew.tsyganov@gmail.com
**************************************************************************************************/


#ifndef SA_OMP_H
#define SA_OMP_H


//#define SA_OMP_DEBUG // Uncomment for debugging


#include <omp.h>
#include <fstream>
#include "sa_base.h"
#include "../../inc/utils/utils.h"


namespace sa
{


/**************************************************************************************************
SA_OMP: abstract class from which user OpenMP algorithm class is derived
**************************************************************************************************/


template <typename TCost, typename TSolution>
class SA_OMP: public SA_BASE < TCost, TSolution >
{
public:
	virtual void run(const SA_Param<TCost>& param, const std::vector<TSolution>& v_init_sol, SA_Result<TCost, TSolution>& res); // Main method
};


template <typename TCost, typename TSolution>
void SA_OMP<TCost, TSolution>::run(const SA_Param<TCost>& param, const std::vector<TSolution>& v_init_sol, SA_Result<TCost, TSolution>& res)
{
	int nthreads = omp_get_max_threads();
	std::vector<omp_lock_t> best_locks;
	bool stop_shared = false;

	std::vector<double> v_init_temp(nthreads);

	res.v_state_.resize(nthreads);
	for (int i = 0; i < nthreads; i++)
	{
		res.v_state_[i].rank_ = i;
		res.v_state_[i].pnum_ = nthreads;
	}

	if (param.comm_rate_ > 0)
	{
		best_locks.resize(nthreads);
		for (int i = 0; i < nthreads; i++)
			omp_init_lock(&best_locks[i]);
	}

	#pragma omp parallel num_threads(nthreads)
	{
		int rank = omp_get_thread_num();
		SA_State<TCost, TSolution>& state = res.v_state_[rank];

		// Begin: obtain ostream

		std::ofstream of;
		std::streambuf* buf = OSBuf(param.ofname_, of, rank, nthreads);
		std::ostream os(buf);

		// End: obtain ostream

		state.start_time_ = get_wall_time();

		int left = (rank - 1 < 0 ? nthreads - 1 : rank - 1); // Determine rank of the left neighbour

		RndGen rnd(seed(rank, nthreads));

		// Begin: set initial temperature

		v_init_temp[rank] = initialTemp(param, state, rnd);

		#pragma omp barrier // Let all threads determine initial temperature

		// Set initial temperature to maximum of all threads
		state.initial_temp_ = v_init_temp[0];
		for (size_t i = 1; i < v_init_temp.size(); i++)
			if (v_init_temp[i] > state.initial_temp_)
				state.initial_temp_ = v_init_temp[i];

		state.temp_ = state.initial_temp_;

		// End: set initial temperature

		// Begin: set initial solution

		if (v_init_sol.size() == 0)
			state.current_ = generate(state, rnd);
		else
		{
			if (v_init_sol.size() >= nthreads)
				state.current_ = v_init_sol[rank];
			else
				state.current_ = v_init_sol[rank % v_init_sol.size()];
		}

		state.current_cost_ = cost(state.current_, state);

		state.best_ = state.current_;
		state.best_cost_ = state.current_cost_;

		// End: set initial solution

		#pragma omp barrier // Let all threads determine initial temperature

		#ifdef SA_OMP_DEBUG
		#pragma omp critical
		std::cout << state << "\n";
		#endif

		output(param, state, os);

		bool stop_priv = stopCondition(param, state);

		if ((param.stop_policy_ == ONE) && stop_priv)
		{
			stop_shared = true;
			#pragma omp flush(stop_shared)
		}

		// Begin: main loop

		while (!(param.stop_policy_ == ONE ? stop_shared : stop_priv))
		{
			state.step_++;
			state.no_change_ = true;

			state.step_iter_ = 0;
			while (state.step_iter_ < param.step_len_)
			{
				state.step_iter_++;

				if ((param.comm_rate_ > 0) && (state.step_ % param.comm_rate_ == 0) && (state.step_iter_ == 1))
				{
					// Begin: communication

					omp_set_lock(&best_locks[left]);

					state.new_ = res.v_state_[left].best_;
					state.new_cost_ = res.v_state_[left].best_cost_;

					omp_unset_lock(&best_locks[left]);

					// End: communication
				}
				else
				{
					state.new_ = neighbour(state.current_, state, rnd);
					state.new_cost_ = cost(state.new_, state);
				}

				if (acceptNew(param, state, rnd))
				{
					state.current_ = state.new_;
					state.current_cost_ = state.new_cost_;

					if (currentBest(param, state))
					{
						if (param.comm_rate_ > 0)
							omp_set_lock(&best_locks[rank]);

						state.no_change_ = false;
						state.best_ = state.current_;
						state.best_cost_ = state.current_cost_;

						if (param.comm_rate_ > 0)
							omp_unset_lock(&best_locks[rank]);
					}
				}

				if (state.step_iter_ == param.step_len_)
					state.stall_step_ = (state.no_change_ ? state.stall_step_ + 1 : 0);

				output(param, state, os);
			}

			#ifdef SA_OMP_DEBUG
			#pragma omp critical
			std::cout << state << "\n";
			#endif

			state.temp_ = newTemp(param, state);

			stop_priv = stopCondition(param, state);

			if ((param.stop_policy_ == ONE) && stop_priv)
			{
				stop_shared = true;
				#pragma omp flush(stop_shared)
			}
		}

		// End: main loop
	}
}


}


#endif
