/**************************************************************************************************
Copyright (C) 2016 Andrey Tsyganov, UlSPU. All rights reserved.

Contact information
-------------------
Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
Web    :  http://lmm.ulspu.ru
e-mail :  andrew.tsyganov@gmail.com
**************************************************************************************************/


#ifndef SHC_BASE_H
#define SHC_BASE_H


//#define SHC_BASE_DEBUG // Uncomment for debugging


#include <vector>
#include <string>
#include "../../contrib/timer/timer.h"
#include "../../contrib/randomc/randomc.h"


namespace shc
{


/**************************************************************************************************
SHC types
**************************************************************************************************/


enum goal_t { MIN, MAX }; // Optimization goal 
enum cond_t { STEPS_LIM, STALL_STEPS_LIM, COST_LIM, TIME_LIM }; // Stopping condition
enum stop_t { ALL, ONE }; // Threads stopping policy


typedef CRandomMersenne RndGen;


/**************************************************************************************************
Forward declarations
**************************************************************************************************/


template <typename TCost>
class SHC_Param;


template <typename TCost>
std::ostream& operator<<(std::ostream& os, const SHC_Param<TCost>& param);


template <typename TCost, typename TSolution>
class SHC_State;


/**************************************************************************************************
SHC_Param: parameters class common for sequential and parallel versions of the SHC algorithm
**************************************************************************************************/


template <typename TCost>
class SHC_Param
{
	template<typename C, typename S> friend class SHC_BASE;
	template<typename C, typename S> friend class SHC_SEQ;
	template<typename C, typename S> friend class SHC_OMP;
	template<typename C, typename S> friend class SHC_MPI;

public:
	// Constructor
	SHC_Param();

	// Parameters setters
	void setGoal(goal_t goal);
	void setStopCond(cond_t stop_cond);
	void setStepsLim(size_t steps_lim);
	void setStallStepsLim(size_t stall_steps_lim);
	void setCostLim(TCost cost_lim);
	void setTimeLim(double time_lim);
	void setStopPolicy(stop_t stop_policy);
	void setCommRate(size_t comm_rate);
	void setOfname(std::string ofname);

	// Parameters getters
	goal_t getGoal() const;
	cond_t getStopCond() const;
	size_t getStepsLim() const;
	size_t getStallStepsLim() const;
	TCost  getCostLim() const;
	double getTimeLim() const;
	stop_t getStopPolicy() const;
	size_t getCommRate() const;
	std::string getOfname() const;

	friend std::ostream& operator<< <>(std::ostream& os, const SHC_Param<TCost>& param);

protected:
	// Parameters
	goal_t goal_;
	cond_t stop_cond_;
	size_t steps_lim_;
	size_t stall_steps_lim_;
	TCost  cost_lim_;
	double time_lim_;
	stop_t stop_policy_;
	size_t comm_rate_;
	std::string ofname_;
};


template <typename TCost>
SHC_Param<TCost>::SHC_Param():
	goal_(MIN),
	stop_cond_(STEPS_LIM),
	steps_lim_(500),
	stall_steps_lim_(100),
	cost_lim_(0),
	time_lim_(1),
	stop_policy_(ALL),
	comm_rate_(0)
{
}


template <typename TCost>
void SHC_Param<TCost>::setGoal(goal_t goal)
{
	goal_ = goal;
}


template <typename TCost>
void SHC_Param<TCost>::setStopCond(cond_t stop_cond)
{
	stop_cond_ = stop_cond;
}


template <typename TCost>
void SHC_Param<TCost>::setStepsLim(size_t steps_lim)
{
	steps_lim_ = steps_lim;
}


template <typename TCost>
void SHC_Param<TCost>::setStallStepsLim(size_t stall_steps_lim)
{
	stall_steps_lim_ = stall_steps_lim;
}


template <typename TCost>
void SHC_Param<TCost>::setCostLim(TCost cost_lim)
{
	cost_lim_ = cost_lim;
}


template <typename TCost>
void SHC_Param<TCost>::setTimeLim(double time_lim)
{
	if (time_lim >= 0)
	{
		time_lim_ = time_lim;
		return;
	};

	#ifdef SHC_BASE_DEBUG
	std::cout << "Incorrect time limit!\n";
	#endif
}


template <typename TCost>
void SHC_Param<TCost>::setStopPolicy(stop_t stop_policy)
{
	stop_policy_ = stop_policy;
}


template <typename TCost>
void SHC_Param<TCost>::setCommRate(size_t comm_rate)
{
	comm_rate_ = comm_rate;
}


template <typename TCost>
void SHC_Param<TCost>::setOfname(std::string ofname)
{
	ofname_ = ofname;
}


template <typename TCost>
goal_t SHC_Param<TCost>::getGoal() const
{
	return goal_;
}

template <typename TCost>
cond_t SHC_Param<TCost>::getStopCond() const
{
	return stop_cond_;
}


template <typename TCost>
size_t SHC_Param<TCost>::getStepsLim() const
{
	return steps_lim_;
}


template <typename TCost>
size_t SHC_Param<TCost>::getStallStepsLim() const
{
	return stall_steps_lim_;
}


template <typename TCost>
TCost SHC_Param<TCost>::getCostLim() const
{
	return cost_lim_;
}


template <typename TCost>
double SHC_Param<TCost>::getTimeLim() const
{
	return time_lim_;
}


template <typename TCost>
stop_t SHC_Param<TCost>::getStopPolicy() const
{
	return stop_policy_;
}


template <typename TCost>
size_t SHC_Param<TCost>::getCommRate() const
{
	return comm_rate_;
}


template <typename TCost>
std::string SHC_Param<TCost>::getOfname() const
{
	return ofname_;
}


template <typename TCost>
std::ostream& operator<<(std::ostream& os, const SHC_Param<TCost>& param)
{
	//enum  { MIN, MAX }; // Optimization goal 
	const char goal_names[2][5] = { "MIN", "MAX" };
	const char cond_names[4][20] = { "STEPS_LIM", "STALL_STEPS_LIM", "COST_LIM", "TIME_LIM" };
	const char stop_names[2][5] = { "ALL", "ONE" };

	os << "goal_ = " << goal_names[param.goal_] << "\n";
	os << "stop_cond_ = " << cond_names[param.stop_cond_] << "\n";
	os << "steps_lim_ = " << param.steps_lim_ << "\n";
	os << "stall_steps_lim_ = " << param.stall_steps_lim_ << "\n";
	os << "cost_lim_ = " << param.cost_lim_ << "\n";
	os << "time_lim_ = " << param.time_lim_ << "\n";
	os << "stop_policy_ = " << stop_names[param.stop_policy_] << "\n";
	os << "comm_rate_ = " << param.comm_rate_ << "\n";
	os << "ofname_ = " << param.ofname_ << "\n";

	return os;
}


/**************************************************************************************************
SHC_State: class for storing current state of the SHC algorithm
**************************************************************************************************/


template <typename TCost, typename TSolution>
class SHC_State
{
public:
	SHC_State();

	int rank_;
	int pnum_;
	double start_time_;
	size_t step_;
	size_t stall_step_;
	bool no_change_;
	TSolution current_;
	TSolution new_;
	TSolution best_;
	TCost current_cost_;
	TCost new_cost_;
	TCost best_cost_;

	friend std::ostream& operator<< <>(std::ostream& os, const SHC_State<TCost, TSolution>& state); // Stream output
};


template <typename TCost, typename TSolution>
SHC_State<TCost, TSolution>::SHC_State():
	rank_(0),
	pnum_(1),
	start_time_(0),
	step_(0),
	stall_step_(0),
	no_change_(false)
{
}


template <typename TCost, typename TSolution>
std::ostream& operator<<(std::ostream& os, const SHC_State<TCost, TSolution>& state)
{
	os << "rank_ = " << state.rank_ << "\n";
	os << "pnum_ = " << state.pnum_ << "\n";
	os << "start_time_ = " << state.start_time_ << "\n";
	os << "step_ = " << state.step_ << "\n";
	os << "stall_step_ = " << state.stall_step_ << "\n";
	os << "no_change_ = " << std::boolalpha << state.no_change_ << "\n";
	os << "current_:\n";
	os << state.current_ << "\n";
	os << "current_cost_ = " << state.current_cost_ << "\n";
	os << "new_:\n";
	os << state.new_ << "\n";
	os << "new_cost_ = " << state.new_cost_ << "\n";
	os << "best_:\n";
	os << state.best_ << "\n";
	os << "best_cost_ = " << state.best_cost_ << "\n";

	return os;
}


/**************************************************************************************************
SHC_Result: class for storing results of the SHC algorithm
**************************************************************************************************/


template <typename TCost, typename TSolution>
class SHC_Result
{
public:
	std::vector< SHC_State<TCost, TSolution> > v_state_;

	friend std::ostream& operator<< <>(std::ostream& os, const SHC_Result<TCost, TSolution>& result); // Stream output
};


template <typename TCost, typename TSolution>
std::ostream& operator<<(std::ostream& os, const SHC_Result<TCost, TSolution>& result)
{
	for (size_t i = 0; i < result.v_state_.size(); i++)
	{
		os << "Thread[" << i << "]:\n";
		os << result.v_state_[i] << "\n";
	}

	return os;
}


/**************************************************************************************************
SHC_BASE: base abstract class from which abstract sequential and parallel algorithms classes
are derived
**************************************************************************************************/


template <typename TCost, typename TSolution>
class SHC_BASE
{
public:
	// These methods may be overriden in derived user classes
	virtual int seed(int rank, int pnum); // Random seed
	virtual void output(const SHC_Param<TCost>& param, const SHC_State<TCost, TSolution>& state, std::ostream& os); // Output current state

	// This method is implemented in derived classes SHC_SEQ, SHC_OMP and SHC_MPI
	virtual void run(const SHC_Param<TCost>& param, const std::vector<TSolution>& v_init_sol, SHC_Result<TCost, TSolution>& res) = 0; // Main method

	// These pure virtual methods are implemented in problem-specific user classes
	virtual TCost cost(const TSolution& solution, const SHC_State<TCost, TSolution>& state) = 0;  // Evaluate cost
	virtual TSolution generate(const SHC_State<TCost, TSolution>& state, RndGen& rnd) = 0; // Generate solution
	virtual TSolution neighbour(const TSolution& solution, const SHC_State<TCost, TSolution>& state, RndGen& rnd) = 0; // Generate neighbour

protected:
	bool stopCondition(const SHC_Param<TCost>& param, const SHC_State<TCost, TSolution>& state); // Stop condition checker
	bool acceptNew(const SHC_Param<TCost>& param, const SHC_State<TCost, TSolution>& state); // Acceptance function
	bool currentBest(const SHC_Param<TCost>& param, const SHC_State<TCost, TSolution>& state); // Check if current is the best solution
};


template <typename TCost, typename TSolution>
int SHC_BASE<TCost, TSolution>::seed(int rank, int pnum)
{
	return (rank << 4) ^ pnum;
}


template <typename TCost, typename TSolution>
void SHC_BASE<TCost, TSolution>::output(const SHC_Param<TCost>& param, const SHC_State<TCost, TSolution>& state, std::ostream& os)
{
}


template <typename TCost, typename TSolution>
bool SHC_BASE<TCost, TSolution>::acceptNew(const SHC_Param<TCost>& param, const SHC_State<TCost, TSolution>& state)
{
	return ((param.goal_ == MIN) && (state.new_cost_ <= state.current_cost_) || (param.goal_ == MAX) && (state.new_cost_ >= state.current_cost_));
}


template <typename TCost, typename TSolution>
bool SHC_BASE<TCost, TSolution>::currentBest(const SHC_Param<TCost>& param, const SHC_State<TCost, TSolution>& state)
{
	return ((param.goal_ == MIN) && (state.current_cost_ < state.best_cost_) || (param.goal_ == MAX) && (state.current_cost_ > state.best_cost_));
}


template <typename TCost, typename TSolution>
bool SHC_BASE<TCost, TSolution>::stopCondition(const SHC_Param<TCost>& param, const SHC_State<TCost, TSolution>& state)
{
	switch (param.stop_cond_)
	{
		case STEPS_LIM:
			return (state.step_ >= param.steps_lim_ ? true : false);

		case STALL_STEPS_LIM:
			return (state.stall_step_ >= param.stall_steps_lim_ ? true : false);

		case COST_LIM:
			if (param.goal_ == MIN)
				return (state.best_cost_ <= param.cost_lim_ ? true : false);
			else
				return (state.best_cost_ >= param.cost_lim_ ? true : false);

		case TIME_LIM:
			return ((get_wall_time() - state.start_time_) >= param.time_lim_ ? true : false);
	}
}


}


#endif
