/**************************************************************************************************
Copyright (C) 2016 Andrey Tsyganov, UlSPU. All rights reserved.

Contact information
-------------------
Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
Web    :  http://lmm.ulspu.ru
e-mail :  andrew.tsyganov@gmail.com
**************************************************************************************************/


#ifndef SHC_SEQ_H
#define SHC_SEQ_H


//#define SHC_SEQ_DEBUG // Uncomment for debugging


#include <ostream>
#include <iostream>
#include <fstream>
#include "shc_base.h"
#include "../../inc/utils/utils.h"


namespace shc
{


/**************************************************************************************************
SHC_SEQ: base abstract class from which user sequential algorithm classes are derived
**************************************************************************************************/


template <typename TCost, typename TSolution>
class SHC_SEQ: public SHC_BASE < TCost, TSolution >
{
public:
	virtual void run(const SHC_Param<TCost>& param, const std::vector<TSolution>& v_init_sol, SHC_Result<TCost, TSolution>& res); // Main method
};


template <typename TCost, typename TSolution>
void SHC_SEQ<TCost, TSolution>::run(const SHC_Param<TCost>& param, const std::vector<TSolution>& v_init_sol, SHC_Result<TCost, TSolution>& res)
{
	SHC_State<TCost, TSolution> state;

	// Begin: obtain ostream

	std::ofstream of;
	std::streambuf* buf = OSBuf(param.ofname_, of, state.rank_, state.pnum_);
	std::ostream os(buf);

	// End: obtain ostream

	state.start_time_ = get_wall_time();

	RndGen rnd(seed(state.rank_, state.pnum_));

	// Begin: set initial solution

	if (v_init_sol.size() == 0)
		state.current_ = generate(state, rnd);
	else
		state.current_ = v_init_sol[0];

	state.current_cost_ = cost(state.current_, state);
	state.best_ = state.current_;
	state.best_cost_ = state.current_cost_;

	#ifdef SHC_SEQ_DEBUG
	std::cout << state << "\n";
	#endif

	// End: set initial solution

	output(param, state, os);

	// Begin: main loop

	while (!stopCondition(param, state))
	{
		state.step_++;
		state.no_change_ = true;

		state.new_ = neighbour(state.current_, state, rnd);
		state.new_cost_ = cost(state.new_, state);

		if (acceptNew(param, state))
		{
			state.current_ = state.new_;
			state.current_cost_ = state.new_cost_;

			if (currentBest(param, state))
			{
				state.no_change_ = false;
				state.best_ = state.current_;
				state.best_cost_ = state.current_cost_;
			}
		}

		state.stall_step_ = (state.no_change_ ? state.stall_step_ + 1 : 0);

		#ifdef SHC_SEQ_DEBUG
		std::cout << state << "\n";
		#endif

		output(param, state, os);
	}

	// End: main loop

	res.v_state_.push_back(state);
}


}


#endif
